///////////////////////////////////////////////////////////////////////////////
// NABLA - a Numerical Analysis Based LAnguage                               //
//                                                                           //
// Copyright (C) 2014~2016 CEA/DAM/DIF                                       //
// IDDN.FR.001.520002.000.S.P.2014.000.10500                                 //
//                                                                           //
// Contributor(s): CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr          //
//                                                                           //
// This software is a computer program whose purpose is to translate         //
// numerical-analysis specific sources and to generate optimized code        //
// for different targets and architectures.                                  //
//                                                                           //
// This software is governed by the CeCILL license under French law and      //
// abiding by the rules of distribution of free software. You can  use,      //
// modify and/or redistribute the software under the terms of the CeCILL     //
// license as circulated by CEA, CNRS and INRIA at the following URL:        //
// "http://www.cecill.info".                                                 //
//                                                                           //
// The CeCILL is a free software license, explicitly compatible with         //
// the GNU GPL.                                                              //
//                                                                           //
// As a counterpart to the access to the source code and rights to copy,     //
// modify and redistribute granted by the license, users are provided only   //
// with a limited warranty and the software's author, the holder of the      //
// economic rights, and the successive licensors have only limited liability.//
//                                                                           //
// In this respect, the user's attention is drawn to the risks associated    //
// with loading, using, modifying and/or developing or reproducing the       //
// software by the user in light of its specific status of free software,    //
// that may mean that it is complicated to manipulate, and that also         //
// therefore means that it is reserved for developers and experienced        //
// professionals having in-depth computer knowledge. Users are therefore     //
// encouraged to load and test the software's suitability as regards their   //
// requirements in conditions enabling the security of their systems and/or  //
// data to be ensured and, more generally, to use and operate it in the      //
// same conditions as regards security.                                      //
//                                                                           //
// The fact that you are presently reading this means that you have had      //
// knowledge of the CeCILL license and that you accept its terms.            //
//                                                                           //
// See the LICENSE file for details.                                         //
///////////////////////////////////////////////////////////////////////////////

// ****************************************************************************
// αβγδεζηθικλμνξοπρςστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ
// ∀ ℝℕℤ ²³√½⅓¼⅛
// ****************************************************************************

// ****************************************************************************
// * Solution exacte, terme Source et Conditions de Dirichlet
// ****************************************************************************


/////////////////////////////////////////////
// Pilotage du Kappa pour faire des Waves
/////////////////////////////////////////////
Real KappaWaves(Real3 p, Real θ){
  if (option_spin_θ)
    return M_PI*sin(option_spin_θ_x*M_PI*p.x)*sin(option_spin_θ_y*M_PI*p.y);
  return θ;
}


/////////////////////////////////////////////
// ISOtropic Diffusion Square withOUT Hole //
/////////////////////////////////////////////
Real exact_solution_isotropic_without_hole(Real3 p){
  return sin(M_PI*p.x)*sin(M_PI*p.y)+option_𝜕Ω_temperature;
}
Real f_isotropic_without_hole(Real3 p){
  return 2.0*M_PI*M_PI*sin(M_PI*p.x)*sin(M_PI*p.y);
}
Real g_isotropic_without_hole(Real3 p){
  return option_𝜕Ω_temperature;
}
Bool is_on_𝜕Ω_isotropic_without_hole(Real3 p){
  // On arrive ici suite à un outer!
  return true;
}


///////////////////////////////////////////////
// ANIsotropic Diffusion Square withOUT Hole //
// But with option_θ that must be tied to 0, //
// for A and B to be positives.              //
///////////////////////////////////////////////
Real exact_solution_anisotropic_without_hole_and_null_θ(Real3 p){
  assert(option_𝜕Ω_temperature==0.0);
  return option_𝜕Ω_temperature+sin(M_PI*p.x)*sin(M_PI*p.y);
}
Real f_anisotropic_without_hole_and_null_θ(Real3 p){
  assert(option_θ==0.0);
  return M_PI*M_PI*((option_k+1.0)*sin(M_PI*p.x)*sin(M_PI*p.y)
    -(option_k-1.0)*cos(M_PI*p.x)*cos(M_PI*p.y)*sin(2.0*option_θ));}
Real g_anisotropic_without_hole_and_null_θ(Real3 p){
  return option_𝜕Ω_temperature;
}
Bool is_on_𝜕Ω_anisotropic_without_hole_and_null_θ(Real3 p){
  // On arrive ici suite à un outer!
  return true;
}


///////////////////////////////////////////////
// ANIsotropic Diffusion Square withOUT Hole //
// Found a function so that p and f are >=0  //
///////////////////////////////////////////////
Real exact_solution_anisotropic_without_hole(Real3 p){
  const Real γ=1.0/atan(½);
  return γ*atan(½-((p.x-½)²+(p.y-½)²));
}
Real f_anisotropic_without_hole(Real3 p){
  const Real x=p.x;
  const Real y=p.y;
  const Real γ=1.0/atan(½);
  const Real θ=option_θ;
  const Real k=option_k;
  const Real dnm = (1.0+(x²-x+y²-y)²)²;
  const Real num =
    -16.0*γ*(k-1.0)*(x-½)*(y-½)*(-x+x²+(y-1.0)*y)*cos(θ)*sin(θ)
    -8.0*γ*(y-½)²*(x²-x+y*(y-1.0))*(k*cos(θ)²+sin(θ)²)
    -8.0*γ*(x-½)²*(x²-x+y*(y-1.0))*(cos(θ)²+k*sin(θ)²)
    +2.0*γ*(1.0+(x²-x+(y-1.0)*y)²)*(k*cos(θ)²+sin(θ)²)
    +2.0*γ*(1.0+(x²-x+(y-1.0)*y)²)*(cos(θ)²+k*sin(θ)²);
  //if (x==0.5 && y==0.5) info()<<"\33[7m[f_anisotropic_without_hole] ("<<x<<","<<y<<")="<<(num/dnm)<<"\33[m";
  return num/dnm;
}
Real g_anisotropic_without_hole(Real3 p){
  return exact_solution_anisotropic_without_hole(p);
}
Bool is_on_𝜕Ω_anisotropic_without_hole(Real3 p){
  // On arrive ici suite à un outer!
  return true;
}


////////////////////////////////////////////
// ANIsotropic Diffusion Square WITH Hole //
////////////////////////////////////////////
Real exact_solution_with_hole(Real3 p){ /*dummy*/ return 1.0;}
Real f_with_hole(Real3 p){ return 0.0;}
Real g_with_hole(Real3 p){
  const Real θ_hole = option_𝜕Ω_temperature+2.0;
  const Real θ_bord = option_𝜕Ω_temperature;
  if (p.x==0.0 || p.x== 1.0) return θ_bord;
  if (p.y==0.0 || p.y== 1.0) return θ_bord;
  return θ_hole;
}
Bool is_on_𝜕Ω_with_hole(Real3 p){
  // On arrive ici suite à un outer!
  return true;
}


////////////////////////////////////////////
// ANIsotropic Diffusion Square WITH Hole //
////////////////////////////////////////////
ℝ u2(ℝ³ p){
  const ℝ x=p.x;
  const ℝ y=p.y;
  //return 1.-2.*y²+4.*x*y+6.*x+2.*y;
  if (x<=0.5) return 1.-2.*y²+4.*x*y+6.*x+2.*y;
  if (x >0.5) return -2.*y²+1.6*x*y-0.6*x+3.2*y+4.3;
  // We should never be there
  assert(false); return 0.0;
}
ℝ exact_solution_gao_wu(ℝ³ p){
  return /*exp(time)**/u2(p);
}
ℝ f_gao_wu(ℝ³ p){
  const ℝ left_value = 4.0;
  const ℝ right_value =
    (option_gao_wu_k==1.0)?-5.6:
    (option_gao_wu_k==2.4)?0.0:
    (option_gao_wu_k==4.0)?6.4:-12345.;
  assert(right_value>-10.0);
  if (p.x < 0.5) return left_value;
  if (p.x== 0.5) return left_value;//Le test sncq pass ici
  if (p.x > 0.5) return right_value;
  assert(false);
  return 0.0;
}
ℝ g_gao_wu(ℝ³ p){
  return exact_solution_gao_wu(p);
}
Bool is_on_𝜕Ω_gao_wu(ℝ³ p){
  // On arrive ici suite à un outer!
  return true;
}



/////////////////////////////////////////////////////////////////////////////
// Switch between Isotropic/Anisotropic Diffusion Square WITH/withOUT Hole //
/////////////////////////////////////////////////////////////////////////////
Real exact_solution(Real3 p){
  if (option_hole) return exact_solution_with_hole(p);
  if (option_isotropic) return exact_solution_isotropic_without_hole(p);
  if (option_atan) return exact_solution_anisotropic_without_hole(p);
  if (option_gao_wu) return exact_solution_gao_wu(p);
  assert(false);
  return exact_solution_anisotropic_without_hole_and_null_θ(p);
}
Real f(Real3 p){
  if (option_hole) return f_with_hole(p);
  if (option_isotropic) return f_isotropic_without_hole(p);
  if (option_atan) return f_anisotropic_without_hole(p);
  if (option_gao_wu) return f_gao_wu(p);
  assert(false);
  return f_anisotropic_without_hole_and_null_θ(p);
}
Real g(Real3 p){
  if (option_hole) return g_with_hole(p);
  if (option_isotropic) return g_isotropic_without_hole(p);
  if (option_atan) return g_anisotropic_without_hole(p);
  if (option_gao_wu) return g_gao_wu(p);
  assert(false);
  return g_anisotropic_without_hole_and_null_θ(p);
}
Bool is_on_𝜕Ω(Real3 p){
  if (option_hole) return is_on_𝜕Ω_with_hole(p);
  if (option_isotropic) return is_on_𝜕Ω_isotropic_without_hole(p);
  if (option_atan) return is_on_𝜕Ω_anisotropic_without_hole(p);
  if (option_gao_wu) return is_on_𝜕Ω_gao_wu(p);
  assert(false);
  return is_on_𝜕Ω_anisotropic_without_hole_and_null_θ(p);
}
