///////////////////////////////////////////////////////////////////////////////
// NABLA - a Numerical Analysis Based LAnguage                               //
//                                                                           //
// Copyright (C) 2014~2016 CEA/DAM/DIF                                       //
// IDDN.FR.001.520002.000.S.P.2014.000.10500                                 //
//                                                                           //
// Contributor(s): CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr          //
//                                                                           //
// This software is a computer program whose purpose is to translate         //
// numerical-analysis specific sources and to generate optimized code        //
// for different targets and architectures.                                  //
//                                                                           //
// This software is governed by the CeCILL license under French law and      //
// abiding by the rules of distribution of free software. You can  use,      //
// modify and/or redistribute the software under the terms of the CeCILL     //
// license as circulated by CEA, CNRS and INRIA at the following URL:        //
// "http://www.cecill.info".                                                 //
//                                                                           //
// The CeCILL is a free software license, explicitly compatible with         //
// the GNU GPL.                                                              //
//                                                                           //
// As a counterpart to the access to the source code and rights to copy,     //
// modify and redistribute granted by the license, users are provided only   //
// with a limited warranty and the software's author, the holder of the      //
// economic rights, and the successive licensors have only limited liability.//
//                                                                           //
// In this respect, the user's attention is drawn to the risks associated    //
// with loading, using, modifying and/or developing or reproducing the       //
// software by the user in light of its specific status of free software,    //
// that may mean that it is complicated to manipulate, and that also         //
// therefore means that it is reserved for developers and experienced        //
// professionals having in-depth computer knowledge. Users are therefore     //
// encouraged to load and test the software's suitability as regards their   //
// requirements in conditions enabling the security of their systems and/or  //
// data to be ensured and, more generally, to use and operate it in the      //
// same conditions as regards security.                                      //
//                                                                           //
// The fact that you are presently reading this means that you have had      //
// knowledge of the CeCILL license and that you accept its terms.            //
//                                                                           //
// See the LICENSE file for details.                                         //
///////////////////////////////////////////////////////////////////////////////

// ****************************************************************************
// * Parties qui s'occupe des inconnues aux faces introduites.
// * Cette façon de faire impose brutalement μ à 0.0 ou 1.0,
// * ceci afin de calculer les valeures aux faces en fonction
// * des valeures voisines aux noeuds et aux mailles
// ****************************************************************************

// On doit pas s'en servir quand on revient au mode average!


// ****************************************************************************
// * Fonction de test du secteur soumis
// ****************************************************************************
Bool iTrialLoopAvg(Bool dbg, Face f, Real3x3 kappa, Real3 νs, Real3 s,
                   Real *pj, Real *pk, Real *α, Real *β, Real *Aire){
  const Real3 D=coord[f->node(0)];
  const Real3 E=coord[f->node(1)];
  const Real3 S=½*(D+E);
  const Real3 P=cell_mass_center[f->backCell()];
  const Real3 Q=cell_mass_center[f->frontCell()];
  //const Integer nid0 = 1+f->node(0)->uniqueId().asInteger();
  //const Integer nid1 = 1+f->node(1)->uniqueId().asInteger();
  Real3 J,K;
  const Real ε=__DBL_EPSILON__;

  // Quart QE
  J=Q;
  K=E;
  *α=round((n(S,K)⋅(kappa⨂νs))/ε)*ε;
  *β=round((n(J,S)⋅(kappa⨂νs))/ε)*ε;
  *pj=cell_θ[f->frontCell()];
  *pk=node_θ[f->node(1)];
  *Aire=geomComputeTriangleArea(S,J,K);
  if (dbg) info()<<"\t\t\33[36m[trialFluxes] QE alpha="<<*α<<", beta="<<*β<<"\33[m";
  if (*α>=0.0 && *β>=0.0) return true;

  // Quart DQ
  J=D;
  K=Q;
  *α=round((n(S,K)⋅(kappa⨂νs))/ε)*ε;
  *β=round((n(J,S)⋅(kappa⨂νs))/ε)*ε;
  *pj=node_θ[f->node(0)];
  *pk=cell_θ[f->frontCell()];
  *Aire=geomComputeTriangleArea(S,J,K);
  if (dbg) info()<<"\t\t\33[36m[trialFluxes] DQ alpha="<<*α<<", beta="<<*β<<"\33[m";
  if (*α>=0.0 && *β>=0.0) return true;

  // Quart PD
  J=P;
  K=D;
  *α=round((n(S,K)⋅(kappa⨂νs))/ε)*ε;
  *β=round((n(J,S)⋅(kappa⨂νs))/ε)*ε;
  *pj=cell_θ[f->backCell()];
  *pk=node_θ[f->node(0)];
  *Aire=geomComputeTriangleArea(S,J,K);
  if (dbg) info()<<"\t\t\33[36m[trialFluxes] PD alpha="<<*α<<", beta="<<*β<<"\33[m";
  if (*α>=0.0 && *β>=0.0) return true;
  
  // Quart EP
  J=E;
  K=P;
  *α=round((n(S,K)⋅(kappa⨂νs))/ε)*ε;
  *β=round((n(J,S)⋅(kappa⨂νs))/ε)*ε;
  *pj=node_θ[f->node(1)];
  *pk=cell_θ[f->backCell()];
  *Aire=geomComputeTriangleArea(S,J,K);
  if (dbg) info()<<"\t\t\33[36m[trialFluxes] EP alpha="<<*α<<", beta="<<*β<<"\33[m";
  if (*α>=0.0 && *β>=0.0) return true;

  return false;
}



// ****************************************************************************
// Pour les faces internes, tout est bien déterminé, pas de ré-orientation à prévoir
// ****************************************************************************
∀ own inner faces void iInnerTrialAvg(void) @ 1.0 if (option_indirect &&
                                                    !option_trial &&                                                    
                                                    !option_trial_average){
  const Real3 D=coord[0];
  const Real3 E=coord[1];
  const Real3 S=½*(D+E);
  const Real3 P=cell_mass_center[backCell];
  const Real3 Q=cell_mass_center[frontCell];
  const Integer nid0 = 1+node(0)->uniqueId().asInteger();
  const Integer nid1 = 1+node(1)->uniqueId().asInteger();
  Real αQE,βQE,AQE,pjQE,pkQE;
  Real αDQ,βDQ,ADQ,pjDQ,pkDQ;
  Real αPD,βPD,APD,pjPD,pkPD;
  Real αEP,βEP,AEP,pjEP,pkEP;
  Bool dbg=option_debug_trial;
  //const Real ε=__DBL_EPSILON__;
  Bool okQE,okDQ,okPD,okEP;
 
  // Pour chaque face en vis-à-vis, on calcul les contributions des diamants
  // Les quatres possibles pour quads sont QE, DQ, PD ou EP
  if (dbg) info()<<"\33[37m[trialFluxes] Face "<<nid0<<"-"<<nid1<<", QE\33[m";
  okQE=iTrialLoopAvg(dbg, this, κ, n(Q,E), S, &pjQE, &pkQE, &αQE, &βQE, &AQE);
  
  if (dbg) info()<<"\33[37m[trialFluxes] Face "<<nid0<<"-"<<nid1<<", DQ\33[m";
  okDQ=iTrialLoopAvg(dbg, this, κ, n(D,Q), S, &pjDQ, &pkDQ, &αDQ, &βDQ, &ADQ);
  
  if (dbg) info()<<"\33[37m[trialFluxes] Face "<<nid0<<"-"<<nid1<<", PD\33[m";
  okPD=iTrialLoopAvg(dbg, this, κ, n(P,D), S, &pjPD, &pkPD, &αPD, &βPD, &APD);
  
  if (dbg) info()<<"\33[37m[trialFluxes] Face "<<nid0<<"-"<<nid1<<", EP\33[m";
  okEP=iTrialLoopAvg(dbg, this, κ, n(E,P), S, &pjEP, &pkEP, &αEP, &βEP, &AEP);
  assert(okQE && okDQ && okPD && okEP);
  if (okQE){
    const Real α=αQE;
    const Real β=βQE;
    const Real A=AQE;
    interior_trial_s[0] = ½*(α+β)/A;
    interior_trial_q[0] = ½*α/A;
    interior_trial_e[0] = ½*β/A;
  }
  if (okDQ){
    const Real α=αDQ;
    const Real β=βDQ;
    const Real A=ADQ;
    interior_trial_s[1] = ½*(α+β)/A;
    interior_trial_q[1] = ½*α/A;
    interior_trial_e[1] = ½*β/A;
  }
  if (okPD){
    const Real α=αPD;
    const Real β=βPD;
    const Real A=APD;
    interior_trial_s[2] = ½*(α+β)/A;
    interior_trial_q[2] = ½*α/A;
    interior_trial_e[2] = ½*β/A;
  }
  if (okEP){
    const Real α=αEP;
    const Real β=βEP;
    const Real A=AEP;
    interior_trial_s[3] = ½*(α+β)/A;
    interior_trial_q[3] = ½*α/A;
    interior_trial_e[3] = ½*β/A;
  }
}




// ****************************************************************************
// * Second membre, coté Trial
// ****************************************************************************
∀ own inner faces void iInnerTrialRhsAvg(void) @ 13.1 if (option_indirect &&
                                                        !option_trial &&
                                                        !option_trial_average){
  const Real3 D=coord[0];
  const Real3 E=coord[1];
  const Real3 S=½*(D+E);
  const Real3 P=cell_mass_center[backCell];
  const Real3 Q=cell_mass_center[frontCell];
  const Real face_area = geomComputeTriangleArea(D,Q,E)+geomComputeTriangleArea(D,E,P);
  const Real h=face_area*f(S);
  const Integer nid0 = 1+node(0)->uniqueId().asInteger();
  const Integer nid1 = 1+node(1)->uniqueId().asInteger();
    
  if (option_debug_trial)
    info()<<"\33[36m[alephRhsTrialInner] Aleph RHS Trial Inner face "
          <<nid0<<"->"<<nid1<<"\33[m";
  ℵ rhs addValue(face_θ,this, h);
}

∀ own outer faces void iOuterTrialRhsAvg(void) @ 13.1 if (!option_trial &&
                                                          option_indirect &&
                                                          !option_trial_average){
  const Real3 D=coord[0];
  const Real3 E=coord[1];
  const Real3 S=½*(D+E);
  const Real h=g(S);
  const Integer nid0 = 1+node(0)->uniqueId().asInteger();
  const Integer nid1 = 1+node(1)->uniqueId().asInteger();
    
  if (option_debug_trial)
    info()<<"\33[36m[alephRhsTrialOuter] Aleph RHS Trial Outer face "
          <<nid0<<"->"<<nid1<<"\33[m";
  ℵ rhs addValue(face_θ,this, h);
 }




// ****************************************************************************
// * Remplissage matriciel, coté Trial
// ****************************************************************************
∀ inner faces void iTrialMtxAvg(void) @ 16.4 if (!option_trial &&
                                               option_indirect &&                                                                                                  
                                               !option_trial_average ){
  const Real svalue =
     interior_trial_s[0]
    +interior_trial_s[1]
    +interior_trial_s[2]
    +interior_trial_s[3];
  const Real q0value = interior_trial_q[0];
  const Real e0value = interior_trial_e[0];
  const Real q1value = interior_trial_q[1];
  const Real e1value = interior_trial_e[1];
  const Real q2value = interior_trial_q[2];
  const Real e2value = interior_trial_e[2];
  const Real q3value = interior_trial_q[3];
  const Real e3value = interior_trial_e[3];
  const Integer nid0 = 1+node(0)->uniqueId().asInteger();
  const Integer nid1 = 1+node(1)->uniqueId().asInteger();
  
  
  if (option_debug_trial)
    info()<<"\t\t\t\33[36m[innerFacesOutside] Aleph Matrix filling face "
          <<nid0<<"->"<<nid1<<"\33[m";
  
  ℵ matrix addValue(face_θ,this, face_θ,this, +svalue);
  
  // Quart QE
  ℵ matrix addValue(face_θ,this, cell_θ,frontCell, -q0value);
  ℵ matrix addValue(face_θ,this, node_θ,node(1),   -e0value);
  
  // Quart DQ
  ℵ matrix addValue(face_θ,this, node_θ,node(0),   -q1value);
  ℵ matrix addValue(face_θ,this, cell_θ,frontCell, -e1value);
  
  // Quart PD
  ℵ matrix addValue(face_θ,this, cell_θ,backCell, -q2value);
  ℵ matrix addValue(face_θ,this, node_θ,node(0),  -e2value);
  
  // Quart EP
  ℵ matrix addValue(face_θ,this, node_θ,node(1),  -q3value);
  ℵ matrix addValue(face_θ,this, cell_θ,backCell, -e3value);
}


// ****************************************************************************
// * Conditions de Dirichlet
// ****************************************************************************
∀ own outer faces void iTrialDirichletAvg(void) @ 17.0 if (!option_trial &&
                                                         option_indirect &&                                                                                                  
                                                        !option_trial_average){
  ℵ matrix addValue(face_θ,this, face_θ,this, +1.0);
}



// ****************************************************************************
// * Récupération des résultats, coté Trial
// ****************************************************************************
∀ own faces void getTrialAvg(void) @ 20.0 if (!option_trial &&
                                            option_indirect &&                                                                                                  
                                            !option_trial_average){
  Real face_θ_kp1;
  face_θ_kp1=ℵ lhs getValue(face_θ,this);
  if (face_θ_kp1<0.0)
    info()<<"\t\33[33m[getFaceResults] face_theta="<< face_θ_kp1 <<"\33[m";
  assert(face_θ_kp1>=0.0);
  if (GlobalIteration==1.0) face_θ_1m0=face_θ_kp1mk;
  face_θ_kp1mk=fabs(face_θ_kp1-face_θ);
  face_θ=face_θ_kp1;
 }

